#include <stdio.h>
#include "erl_nif.h"

#define DEBUG

#ifdef DEBUG
#include <stdarg.h>
static void emit_error(char* file, int line, ...);
#define DBG(...) emit_error(__FILE__,__LINE__,__VA_ARGS__)
#else
#define DBG(...)
#endif

typedef struct {
    int n;
} niffy_t;

static int niffy_load(ErlNifEnv* env, void** priv_data, ERL_NIF_TERM load_info);

static int niffy_upgrade(ErlNifEnv* env, void** priv_data, void** old_priv_data, 
			 ERL_NIF_TERM load_info);
static void niffy_unload(ErlNifEnv* env, void* priv_data);

static ERL_NIF_TERM niffy_hello(ErlNifEnv* env, int argc, 
				const ERL_NIF_TERM argv[]);

ErlNifFunc niffy_funcs[] =
{
    { "hello",  0, niffy_hello }
};

static void emit_error(char* file, int line, ...)
{
    va_list ap;
    char* fmt;

    va_start(ap, line);
    fmt = va_arg(ap, char*);

    fprintf(stderr, "%s:%d: ", file, line); 
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\r\n");
    va_end(ap);
    fflush(stderr);
}

static ERL_NIF_TERM niffy_hello(ErlNifEnv* env, int argc, 
				const ERL_NIF_TERM argv[])
{
    niffy_t* ptr = enif_priv_data(env);
    DBG("niffy_hello: priv_data=%p", ptr);
    return enif_make_atom(env, "world");
}

static int niffy_load(ErlNifEnv* env, void** priv_data, ERL_NIF_TERM load_info)
{
    niffy_t* ptr;
    ptr = enif_alloc(sizeof(niffy_t));
    ptr->n = 1;
    DBG("niffy_load: *priv_data=%p ptr=%p, ptr->n", 
	*priv_data, ptr, ptr->n);
    *priv_data = ptr;
    return 0;
}

static int niffy_upgrade(ErlNifEnv* env, void** priv_data, void** old_priv_data,
			 ERL_NIF_TERM load_info)
{
    niffy_t* ptr = *old_priv_data;

    DBG("niffy_upgrade: *priv_data=%p *old_priv_data=%p ptr->n=%d",
	*priv_data, *old_priv_data, ptr->n);
    ptr->n++;
    *priv_data = *old_priv_data;
    return 0;
}

static void niffy_unload(ErlNifEnv* env, void* priv_data)
{
    niffy_t* ptr = priv_data;
    DBG("niffy_unload: priv_data=%p ptr->n=%d", priv_data, ptr->n);
    enif_free(ptr);
}

ERL_NIF_INIT(niffy, niffy_funcs, 
	     niffy_load, NULL,
	     niffy_upgrade, niffy_unload)
